#ifndef __GRAPH_H__
#define __GRAPH_H__

#include "Array.h"
#include "Array2D.h"
#include "Edge.h"
#include "Vertex.h"

#include <iostream>
#include <queue>
#include "PriorityQueue.h"
#include "Dictionary.h"

const int DIST_INF = 9999;



class Graph {
    Array<Vertex> _vertices;
    void(*_action)(int v);

    Array2D<int> *_length;

public:
    Graph(int vertices_count);
    ~Graph() {
        delete _length;
    }

    void setAction(void(*action)(int v));



    void dilink(int v1, int v2, int cost) {
        _vertices[v1].adjacencyList().push_back(Edge(v1, v2, cost));
    }
    void link(int v1, int v2, int cost);



    void dfs(int v) const;
    void bfs(int v) const;

    void Components() const;
    void OutputNewComponent(const Array<bool> &visited) const;

    void kruskal() const;
    void prim() const;



    void init_length() {
        const int n = _vertices.count();
        _length = new Array2D<int>(n, n);
        for (int r = 0; r < n; ++r) {
            for (int c = 0; c < n; ++c) {
                if (r == c)
                    _length->set(0, r, c);
                else
                    _length->set(DIST_INF, r, c);
            }
        }
        for (int i = 0; i < n; ++i) {
            const std::vector<Edge> &adjList = _vertices[i].adjacencyList();
            for (int j = 0, adjCount = adjList.size(); j < adjCount; ++j) {
                const Edge &edge = adjList[j];
                _length->set(edge.cost(), edge.source(), edge.destination());
            }
        }
    }
    void dijkstra(int v) const {
        const int n = _vertices.count();
        const Array2D<int> &length = *_length;
        Array<bool> sp_found(n);
        Array<int> distance(n);

        for (int i = 0; i < n; ++i) {
            sp_found[i] = false;
            distance[i] = length.at(v, i);
        }
        sp_found[v] = true;

        for (int i = 0; i < n - 2; ++i) {
            int u = choose_min_dist(distance, sp_found);
            if (u < 0)
                break; // dijkstra end
            sp_found[u] = true;

            for (int w = 0; w < n; ++w) {
                if (sp_found[w] == false
                    && distance[w] > distance[u] + length.at(u, w)) {
                    distance[w] = distance[u] + length.at(u, w);
                }
            }
        }

        for (int i = 0; i < n; ++i) {
            std::cout << '\t' << distance[i];
        }
        std::cout << std::endl;
    }
    int choose_min_dist(const Array<int> &dist, const Array<bool> &sp_found) const {
        const int n = dist.count();
        int min_cost = INT_MAX, min_idx = -1;
        for (int i = 0; i < n; ++i) {
            if (sp_found[i] == false && dist[i] < min_cost) {
                min_cost = dist[i];
                min_idx = i;
            }
        }
        return min_idx;
    }



private:
    void _dfs(int v, Array<bool> &visited) const;
    void _bfs(int v, Array<bool> &visited) const;
    void _visit(int v, Array<bool> &visited) const;

    bool _path_exist(int src, int dst) const;
};

#endif