//////////////////////////////////////////////////////////////
// File Name : Vector.h
// Date : 2015/12/04
// Compiler : g++ 4.6.3
// Os : ubuntu 12.04.5
// Author : Do Yeong Han
// Student ID : 2011722086
// =========================================================
// ver : 2.1.7
// Description : Vector data structure
// Etc.. : Data Structure Assignment #3
//  ( Link State Routing Simulation )
//////////////////////////////////////////////////////////////
#ifndef __HANDY_VECTOR_H__
#define __HANDY_VECTOR_H__

#include "handy.h"



template <class T>
class Vector {
    ////////////////////////////////////////////////////////////////////
    // auxilary class definition
    class VectorNode {
        T _data;
        VectorNode *_next;

    public:
        ////////////////////////////////////////////////////////////////////
        // VectorNode::VectorNode
        // =================================================================
        // Input:
        //  none
        // Purpose: 
        //  constructor
        ////////////////////////////////////////////////////////////////////
        VectorNode() : _data(), _next(null) {

        }
        ////////////////////////////////////////////////////////////////////
        // VectorNode::VectorNode
        // =================================================================
        // Input:
        //  (const T& data) -> new data of node
        // Purpose: 
        //  constructor
        ////////////////////////////////////////////////////////////////////
        VectorNode(const T& data) : _data(data), _next(null) {

        }
        ////////////////////////////////////////////////////////////////////
        // VectorNode::~VectorNode
        // =================================================================
        // Purpose: 
        //  constructor
        ////////////////////////////////////////////////////////////////////
        ~VectorNode() {

        }



        ////////////////////////////////////////////////////////////////////
        // VectorNode::data
        // =================================================================
        // Input:
        //  none
        // Output:
        //  (T&) -> reference of node's data
        // Purpose: 
        //  getter method for field _data
        ////////////////////////////////////////////////////////////////////
        T& data() {
            return _data;
        }
        ////////////////////////////////////////////////////////////////////
        // VectorNode::data
        // =================================================================
        // Input:
        //  none
        // Output:
        //  (const T&) -> read-only reference of node's data
        // Purpose: 
        //  getter method for field _data
        ////////////////////////////////////////////////////////////////////
        const T& data() const {
            return _data;
        }
        ////////////////////////////////////////////////////////////////////
        // VectorNode::setData
        // =================================================================
        // Input:
        //  (const T& value) -> new data value to update node's data
        // Output:
        //  none
        // Purpose: 
        //  setter method for field _data
        ////////////////////////////////////////////////////////////////////
        void setData(const T& value) {
            _data = value;
        }
        ////////////////////////////////////////////////////////////////////
        // VectorNode::next
        // =================================================================
        // Input:
        //  none
        // Output:
        //  (VectorNode *) -> next node's address
        // Purpose: 
        //  getter method for field _next
        ////////////////////////////////////////////////////////////////////
        VectorNode *next() const {
            return _next;
        }
        ////////////////////////////////////////////////////////////////////
        // VectorNode::setNext
        // =================================================================
        // Input:
        //  (VectorNode *value) -> new node address to update node's field
        // Output:
        //  none
        // Purpose: 
        //  setter method for field _next
        ////////////////////////////////////////////////////////////////////
        void setNext(VectorNode *value) {
            _next = value;
        }
    };

    // end of auxilary class definition
    ////////////////////////////////////////////////////////////////////



    ////////////////////////////////////////////////////////////////////
    // field definition
    VectorNode _head;   // head node pointer (dummy)
    VectorNode *_tail;  // tail node pointer
    int _count;         // element count in vector

    // end of field definition
    ////////////////////////////////////////////////////////////////////



public:
    ////////////////////////////////////////////////////////////////////
    // Vector::Vector
    // =================================================================
    // Input:
    //  none
    // Purpose: 
    //  constructor
    ////////////////////////////////////////////////////////////////////
    Vector() : _head(), _count(0) {
        _tail = &_head; // initialize tail pointer to head pointer's address
    }
    ////////////////////////////////////////////////////////////////////
    // Vector::Vector
    // =================================================================
    // Input:
    //  (const Vector& copy) -> reference to vector to copy
    // Purpose: 
    //  copy constructor
    ////////////////////////////////////////////////////////////////////
    Vector(const Vector& copy): _head(), _count(0) {
        int count = copy.count(); // get element count of vector to copy
        VectorNode *tail = &_head; // initialize tail pointer
        VectorNode *newNode = null; // for newly-allocated node

        for (int i = 0; i < count; ++i) { // loop for "copy element"
            newNode = new VectorNode(copy[i]); // create new node
            tail->setNext(newNode); // link vector and new node
            tail = newNode; // update tail pointer to new node
        } // end of loop "copy element"

        // initialize fields
        _tail = tail;
        _count = count;
    }
    ////////////////////////////////////////////////////////////////////
    // Vector::~Vector
    // =================================================================
    // Purpose: 
    //  destructor
    ////////////////////////////////////////////////////////////////////
    ~Vector() {
        /// deallocate vector's element ////////////////////////////////
        VectorNode *del_node = _head.next(), *del_next = null;
        while (del_node) { // loop while "deallocate"
            del_next = del_node->next(); // point next node first
            delete del_node; // deallocate present node
            del_node = del_next; // update walker to next
        } // end of loop "deallocate"

        /// end of deallocation ////////////////////////////////////////
    }



    ////////////////////////////////////////////////////////////////////
    // Vector::empty
    // =================================================================
    // Input:
    //  none
    // Output:
    //  (bool) -> true when vector is empty; false when not
    // Purpose: 
    //  check whether vector is empty
    ////////////////////////////////////////////////////////////////////
    bool empty() const {
        return _count == 0; // vector is empty when it's count equals to 0
    }
    ////////////////////////////////////////////////////////////////////
    // Vector::count
    // =================================================================
    // Input:
    //  none
    // Output:
    //  (int) -> element count in vector (empty vector returns 0)
    // Purpose: 
    //  getter method for field _count
    ////////////////////////////////////////////////////////////////////
    int count() const {
        return _count;
    }



    ////////////////////////////////////////////////////////////////////
    // Vector::operator []
    // =================================================================
    // Input:
    //  (int index) -> element index to access
    // Output:
    //  (T&) -> reference to element
    // Purpose: 
    //  access to element with passed index
    ////////////////////////////////////////////////////////////////////
    T& operator [](int index) {
        /// access to element in vector ////////////////////////////////
        if (out_of_range(index)) // index is NOT in range; throw exception
            throw Exception("cannot access to that index");

        VectorNode *node = _head.next(); // start from dummy's next
        for (int i = 0; i < index; ++i) // loop until i equals to index
            node = node->next(); // move walker to next node
        return node->data(); // return element's data

        /// end of access //////////////////////////////////////////////
    }
    ////////////////////////////////////////////////////////////////////
    // Vector::operator []
    // =================================================================
    // Input:
    //  (int index) -> element index to access
    // Output:
    //  (const T&) -> read-only reference to element
    // Purpose: 
    //  access to element with passed index
    ////////////////////////////////////////////////////////////////////
    const T& operator [](int index) const {
        /// access to element in vector ////////////////////////////////
        if (out_of_range(index)) // index is NOT in range; throw exception
            throw Exception("cannot access to that index");

        VectorNode *node = _head.next(); // start from dummy's next
        for (int i = 0; i < index; ++i) // loop until i equals to index
            node = node->next(); // move walker to next node
        return node->data(); // return element's data

        /// end of access //////////////////////////////////////////////
    }
    ////////////////////////////////////////////////////////////////////
    // Vector::exist
    // =================================================================
    // Input:
    //  (const T& target) -> data to check whether is in vector
    // Output:
    //  (bool) -> true when target exists, false when not
    // Purpose: 
    //  check element exists in list
    ////////////////////////////////////////////////////////////////////
    bool exist(const T& target) const {
        /// find out the target is in vector ///////////////////////////
        if (empty()) // vector is empty
            return false; // element do not exists in empty vector

        VectorNode *node = _head.next(); // start from dummy's next
        while (node) { // loop while "finding target"; while node exists
            if (node->data() == target) // target was found
                return true; // target exists in vector
            node = node->next(); // move to next node
        } // end of loop "finding target"
        return false; // cannot found target; target does NOT exist in vector

        /// end of finding target //////////////////////////////////////
    }



    ////////////////////////////////////////////////////////////////////
    // Vector::insert
    // =================================================================
    // Input:
    //  (const T& data) -> data to insert into vector
    // Output:
    //  none
    // Purpose: 
    //  insert data into vector
    ////////////////////////////////////////////////////////////////////
    void insert(const T& data) {
        /// push back new data into vector /////////////////////////////
        VectorNode *newNode = new VectorNode(); // create new node
        newNode->setData(data); // initialize it with parameter

        VectorNode *prev = &_head, *node = _head.next();
        while (node) { // loop while "finding position of new node"
            if (data <= node->data()) // newly-inserted node's position found
                break; // exit loop
            prev = node; // update previous node
            node = node->next(); // move node pointer to next node
        } // end of loop "finding position of new node"

        if (node) { // next node exists
            // insert new node between prev and node
            newNode->setNext(node); // link new node and next node
            prev->setNext(newNode); // link previous node and new node
        }
        else { // reached end of vector
            _tail->setNext(newNode); // link created node and vector
            _tail = newNode; // update tail pointer to newly-created node
        }
        ++_count; // increase vector count

        /// end of node insertion //////////////////////////////////////
    }



    ////////////////////////////////////////////////////////////////////
    // Vector::push_back
    // =================================================================
    // Input:
    //  (const T& data) -> new data to insert
    // Output:
    //  none
    // Purpose: 
    //  push back new data into vector
    ////////////////////////////////////////////////////////////////////
    void push_back(const T& data) {
        /// push back new data into vector /////////////////////////////
        VectorNode *newNode = new VectorNode(); // create new node
        newNode->setData(data); // initialize it with parameter

        _tail->setNext(newNode); // link created node and vector
        _tail = newNode; // update tail pointer to newly-created node
        ++_count; // increase vector count

        /// end of node insertion //////////////////////////////////////
    }
    ////////////////////////////////////////////////////////////////////
    // Vector::remove
    // =================================================================
    // Input:
    //  (int index) -> element index to delete from vector
    // Output:
    //  none
    // Purpose: 
    //  remove element at index from vector
    ////////////////////////////////////////////////////////////////////
    void VectorRemoveCall(int index) {
        /// remove element at index from vector ////////////////////////
        if (out_of_range(index)) // invalid index has entered
            return; // nothing to do; ignore message

        VectorNode *prev = &_head, *node = _head.next(); // start from dummy's first
        for (int i = 0; i < index; ++i) { // loop 'index' times
            prev = node; // update previous pointer to walker
            node = node->next(); // update walker to next
        }
        prev->setNext(node->next()); // link walker's previous node and next node
        delete node; // deallocate target
        --_count; // decrease vector count

        if (index == _count) // last element removed
            _tail = prev; // re-position tail pointer to previous node

        /// end of element remove //////////////////////////////////////
    }



    ////////////////////////////////////////////////////////////////////
    // Vector::find
    // =================================================================
    // Input:
    //  (const T& target) -> data to check whether is in vector
    // Output:
    //  (int) -> index of target in vector
    //           -1 when there's NOT target in vector
    // Purpose: 
    //  check element exists in list
    ////////////////////////////////////////////////////////////////////
    int find(const T& target) const {
        /// find out the target is in vector ///////////////////////////
        if (empty()) // vector is empty
            return -1; // element do not exists in empty vector

        VectorNode *node = _head.next(); // start from dummy's next
        for (int i = 0; node; ++i) { // loop while "finding target"; while node exists
            if (node->data() == target) // target was found
                return i; // target exists in vector
            node = node->next(); // move to next node
        } // end of loop "finding target"
        return -1; // cannot found target; target does NOT exist in vector

        /// end of finding target //////////////////////////////////////
    }
    ////////////////////////////////////////////////////////////////////
    // Vector::traverse
    // =================================================================
    // Input:
    //  (void (*Action)(const T&)) -> action for each node's data
    // Output:
    //  none
    // Purpose: 
    //  traverse each nodes in vector sequentially
    ////////////////////////////////////////////////////////////////////
    void traverse(void(*Action)(const T&)) const {
        /// traverse each nodes in vector sequentially /////////////////
        VectorNode *node = _head.next(); // starts from first node
        while (node) { // loop while "do action"
            Action(node->data()); // do action for each node's data
            node = node->next(); // move node pointer to next
        } // end of loop "do action"
        
        /// end of node traversing /////////////////////////////////////
    }



    ////////////////////////////////////////////////////////////////////
    // Vector::operator =
    // =================================================================
    // Input:
    //  (int index) -> element index to delete from vector
    // Output:
    //  none
    // Purpose: 
    //  overriding for substitute operator
    //  (also overloading; overriding means, there's always
    //   default substitute operator when it is NOT defined by developer)
    ////////////////////////////////////////////////////////////////////
    Vector& operator =(const Vector& copy) {
        /// deallocate previous elements ///////////////////////////////
        if (empty() == false) { // clear element left in vector
            VectorNode *del_node = _head.next(), *del_next;
            while (del_node) { // loop while "deallocate previous element"
                del_next = del_node->next(); // preserve next node pointer first
                delete del_node; // deallocate present node
                del_node = del_next; // update walker to next pointer
            } // end of loop "deallocate previous element"
        }

        /// end of clear vector ////////////////////////////////////////



        /// copy vector ////////////////////////////////////////////////
        VectorNode *tail = &_head, *new_node;
        int count = copy.count(); // element count to copy
        for (int i = 0; i < count; ++i) { // loop for "copy element"
            new_node = new VectorNode(copy[i]); // create new node
            tail->setNext(new_node); // link vector and new node
            tail = new_node; // update tail pointer to new node
        } // end of loop "copy element"

        // initialize field
        _tail = tail;
        _count = count;

        /// end of copying /////////////////////////////////////////////

        return (*this); // return self (conventially)
    }



private:
    ////////////////////////////////////////////////////////////////////
    // Vector::out_of_range
    // =================================================================
    // Input:
    //  (int index) -> index to check whether it is out of range
    // Output:
    //  (bool) -> true when index is invalid (less than 0 or >= than count)
    //            false: index is good to use (in valid range)
    // Purpose: 
    //  check whether index is out of range
    ////////////////////////////////////////////////////////////////////
    bool out_of_range(int index) const {
        return (index < 0 || _count <= index);
    }
};

#endif
