#ifndef __HANDY_ARRAY2D_H__
#define __HANDY_ARRAY2D_H__

#include "handy.h"



template <class T>
class Array2D {
    T *_list;
    int _row;
    int _column;

public:
    Array2D(int row, int col) : _row(row), _column(col) {
        _list = new T[row * col];
    }
    Array2D(const Array2D& copy) {
        int row = copy.row(), col = copy.column();
        _row = row;
        _column = col;

        _list = new T[row * col];
        for (int r = 0; r < row; ++r) {
            for (int c = 0; c < col; ++c) {
                set(copy.at(r, c), r, c);
            }
        }
    }
    ~Array2D() {
        delete[] _list;
    }

    int row() const { return _row; }
    int column() const { return _column; }

    T &at(int r, int c) {
        return _list[r * _column + c];
    }
    const T &at(int r, int c) const {
        return _list[r * _column + c];
    }
    void set(const T& value, int r, int c) {
        _list[r * _column + c] = value;
    }

    Array2D &operator =(const Array2D &copy) {
        if (_list)
            delete[] _list;
        int row = copy.row(), col = copy.column();
        _list = new T[row * col];
        for (int r = 0; r < row; ++r) {
            for (int c = 0; c < col; ++c) {
                _list.set(copy.at(r, c), r, c);
            }
        }
        _row = row;
        _column = col;
        return (*this);
    }
    bool operator ==(const Array2D &o) const {
        if (_row != o.row() || _column != o.column())
            return false;

        for (int r = 0; r < _row; ++r) {
            for (int c = 0; c < _column; ++c) {
                if (at(r, c) != o.at(r, c))
                    return false;
            }
        }

        return true;
    }
};

#endif