//////////////////////////////////////////////////////////////
// File Name : Dictionary.h
// Date : 2015/12/02
// Compiler : g++ 4.6.3
// Os : ubuntu 12.04.5
// Author : Do Yeong Han
// Student ID : 2011722086
// =========================================================
// ver : 2.0.7
// Description : Dictionary data structure
// Etc.. : Data Structure Assignment #3
//  ( Link State Routing Simulation )
//////////////////////////////////////////////////////////////
#ifndef __HANDY_DICTIONARY_H__
#define __HANDY_DICTIONARY_H__

#include "handy.h"
#include "Vector.h"
// #include "PriorityQueue.h"



template <class Key, class Value>
class Dictionary {
    ////////////////////////////////////////////////////////////////////
    // auxilary class definition
    class Pair {
        Key _key;
        Value _value;
    public:
        ////////////////////////////////////////////////////////////////////
        // Pair::Pair
        // =================================================================
        // Input:
        //  none
        // Purpose: 
        //  constructor
        ////////////////////////////////////////////////////////////////////
        Pair() : _key(), _value() { }
        ////////////////////////////////////////////////////////////////////
        // Pair::Pair
        // =================================================================
        // Input:
        //  (const Key& key) -> 
        // Purpose: 
        //  constructor
        ////////////////////////////////////////////////////////////////////
        Pair(const Key& key) : _key(key), _value() { }
        ////////////////////////////////////////////////////////////////////
        // Pair::Pair
        // =================================================================
        // Input:
        //  (const Key& key) -> 
        //  (const Value& value) -> 
        // Purpose: 
        //  constructor
        ////////////////////////////////////////////////////////////////////
        Pair(const Key& key, const Value& value) : _key(key), _value(value) { }
        ////////////////////////////////////////////////////////////////////
        // Pair::~Pair
        // =================================================================
        // Input:
        //  none
        // Purpose: 
        //  destructor
        ////////////////////////////////////////////////////////////////////
        ~Pair() { }



        ////////////////////////////////////////////////////////////////////
        // Pair::key
        // =================================================================
        // Input:
        //  none
        // Output:
        //  (Key &) -> reference of field _key
        // Purpose: 
        //  getter method for field _key
        ////////////////////////////////////////////////////////////////////
        Key &key() {
            return _key;
        }
        ////////////////////////////////////////////////////////////////////
        // Pair::key
        // =================================================================
        // Input:
        //  none
        // Output:
        //  (const Key&) -> constant reference of field _key
        // Purpose: 
        //  getter method for field _key
        ////////////////////////////////////////////////////////////////////
        const Key &key() const {
            return _key;
        }
        ////////////////////////////////////////////////////////////////////
        // Pair::setKey
        // =================================================================
        // Input:
        //  (const Key& key) -> new key to update
        // Output:
        //  none
        // Purpose: 
        //  setter method for field _key
        ////////////////////////////////////////////////////////////////////
        void setKey(const Key& key) {
            _key = key;
        }
        ////////////////////////////////////////////////////////////////////
        // Pair::value
        // =================================================================
        // Input:
        //  none
        // Output:
        //  (Value &) -> reference of field _value
        // Purpose: 
        //  getter method for field _value
        ////////////////////////////////////////////////////////////////////
        Value &value() {
            return _value;
        }
        ////////////////////////////////////////////////////////////////////
        // Pair::value
        // =================================================================
        // Input:
        //  none
        // Output:
        //  (const Value&) -> constant reference of field _value
        // Purpose: 
        //  getter method for field _value
        ////////////////////////////////////////////////////////////////////
        const Value &value() const {
            return _value;
        }
        ////////////////////////////////////////////////////////////////////
        // Pair::setValue
        // =================================================================
        // Input:
        //  (const Value& value) -> new value to update
        // Output:
        //  none
        // Purpose: 
        //  setter method for field _value
        ////////////////////////////////////////////////////////////////////
        void setValue(const Value& value) {
            _value = value;
        }



        ////////////////////////////////////////////////////////////////////
        // Pair::operator ==
        // =================================================================
        // Input:
        //  (const Pair& pair) -> object to compare
        // Output:
        //  (bool) -> 
        // Purpose: 
        //  comparer; check both object's keys are equal or not
        ////////////////////////////////////////////////////////////////////
        bool operator ==(const Pair& pair) const {
            return _key == pair.key();
        }
        bool operator <=(const Pair& pair) const {
            return _key <= pair.key();
        }
    };

    // end of auxilary class definition
    ////////////////////////////////////////////////////////////////////



    ////////////////////////////////////////////////////////////////////
    // field definition
    Vector<Pair> _dict;

    // end of field definition
    ////////////////////////////////////////////////////////////////////



public:
    ////////////////////////////////////////////////////////////////////
    // Dictionary::Dictionary
    // =================================================================
    // Input:
    //  none
    // Purpose: 
    //  constructor
    ////////////////////////////////////////////////////////////////////
    Dictionary() : _dict() {
        
    }
    ////////////////////////////////////////////////////////////////////
    // Dictionary::~Dictionary
    // =================================================================
    // Purpose: 
    //  destructor
    ////////////////////////////////////////////////////////////////////
    ~Dictionary() {

    }



    ////////////////////////////////////////////////////////////////////
    // Dictionary::empty
    // =================================================================
    // Input:
    //  none
    // Output:
    //  (bool) -> true when dictionary has no elements; false when not
    // Purpose: 
    //  check whether dictionary is empty
    ////////////////////////////////////////////////////////////////////
    bool empty() const {
        return _dict.empty();
    }
    ////////////////////////////////////////////////////////////////////
    // Dictionary::exist
    // =================================================================
    // Input:
    //  (const Key& key) -> target's key to check
    // Output:
    //  (bool) -> true when target is in dictionary; false when not
    // Purpose: 
    //  check whether element exists in dictionary
    ////////////////////////////////////////////////////////////////////
    bool exist(const Key& key) const {
        return _dict.exist(Pair(key));
    }
    ////////////////////////////////////////////////////////////////////
    // Dictionary::count
    // =================================================================
    // Input:
    //  none
    // Output:
    //  (int) -> element count in dictionary
    // Purpose: 
    //  get element count in dictionary
    ////////////////////////////////////////////////////////////////////
    int count() const {
        return _dict.count();
    }



    ////////////////////////////////////////////////////////////////////
    // Dictionary::set
    // =================================================================
    // Input:
    //  (const Key& key) -> new element's key
    //  (const Value& value) -> new element's value
    // Output:
    //  none
    // Purpose: 
    //  set value for key
    ////////////////////////////////////////////////////////////////////
    void set(const Key& key, const Value& value) {
        /// set value for key //////////////////////////////////////////
        if (exist(key)) // key already exists in dictionary
            (*this)[key] = value; // update value for key
        else // key is NOT included in dictionary yet
            _dict.insert(Pair(key, value)); // insert key-value pair object

        /// end of setting value for key ///////////////////////////////
    }



    ////////////////////////////////////////////////////////////////////
    // Dictionary::remove
    // =================================================================
    // Input:
    //  (const Key& key) -> element's key to remove
    // Output:
    //  none
    // Purpose: 
    //  remove value for key from dictionary
    ////////////////////////////////////////////////////////////////////
    void DictionaryRemoveCall(const Key& key) {
        _dict.VectorRemoveCall(_dict.find(key));
    }



    ////////////////////////////////////////////////////////////////////
    // Dictionary::atIndex
    // =================================================================
    // Input:
    //  (int index) -> element index to access
    // Output:
    //  (Value &) -> reference to value of element at index
    // Purpose: 
    //  get reference to element value at index
    ////////////////////////////////////////////////////////////////////
    Value &atIndex(int index) {
        return _dict[index].value();
    }
    ////////////////////////////////////////////////////////////////////
    // Dictionary::atIndex
    // =================================================================
    // Input:
    //  (int index) -> element index to access
    // Output:
    //  (const Value &) -> constant reference to value of element at index
    // Purpose: 
    //  get constant reference to element value at index
    ////////////////////////////////////////////////////////////////////
    const Value &atIndex(int index) const {
        return _dict[index].value();
    }



    ////////////////////////////////////////////////////////////////////
    // Dictionary::operator []
    // =================================================================
    // Input:
    //  (const Key& key) -> element's key to find
    // Output:
    //  (Value &) -> element's value found by key
    // Purpose: 
    //  access to dictionary element
    ////////////////////////////////////////////////////////////////////
    Value &operator [](const Key& key) {
        /// access to dictionary element ///////////////////////////////
        for (int i = 0, len = _dict.count(); i < len; ++i) { // loop for "finding value for key"
            Pair& pair = _dict[i]; // pair object to compare
            if (pair.key() == key) // value that matches key was found
                return pair.value(); // return value for key
        } // end of for "finding value for key"
        throw Exception("cannot find target"); // cannot find element for key

        /// end of element access //////////////////////////////////////
    }
    ////////////////////////////////////////////////////////////////////
    // Dictionary::operator []
    // =================================================================
    // Input:
    //  (const Key& key) -> element's key to find
    // Output:
    //  (const Value &) -> element's value found by key
    // Purpose: 
    //  access to dictionary element
    ////////////////////////////////////////////////////////////////////
    const Value &operator [](const Key& key) const {
        /// access to dictionary element ///////////////////////////////
        for (int i = 0, len = _dict.count(); i < len; ++i) { // loop for "finding value for key"
            const Pair& pair = _dict[i]; // pair object to compare
            if (pair.key() == key) // value that matches key was found
                return pair.value(); // return value for key
        } // end of for "finding value for key"
        throw Exception("cannot find target"); // cannot find element for key

        /// end of element access //////////////////////////////////////
    }
};

#endif