#include "Graph.h"



Graph::Graph(int vertices_count) : _vertices(vertices_count), _action(nullptr), _length(nullptr) {

}
Graph::~Graph() {
    delete _length;
}


void Graph::setAction(void(*action)(int v)) {
    _action = action;
}



void Graph::dilink(int v1, int v2, int cost) {
    _vertices[v1].adjacencyList().push_back(Edge(v1, v2, cost));
}
void Graph::link(int v1, int v2, int cost) {
    _vertices[v1].adjacencyList().push_back(Edge(v1, v2, cost));
    _vertices[v2].adjacencyList().push_back(Edge(v2, v1, cost));
}



void Graph::dfs(int v) const { // driver
    const int n = _vertices.count();
    Array<bool> visited(n);
    for (int i = 0; i < n; ++i) {
        visited[i] = false;
    }
    _dfs(v, visited);
}
void Graph::bfs(int v) const {
    const int n = _vertices.count();
    Array<bool> visited(n);
    for (int i = 0; i < n; ++i)
        visited[i] = false;
    _bfs(v, visited);
}



void Graph::Components() const {
    // determine the connected components of the graph.
    // visited is assumed to be declared as a bool * data member of Graph
    const int n = _vertices.count();
    Array<bool> visited(n);
    for (int i = 0; i < n; ++i)
        visited[i] = false;

    for (int i = 0; i < n; ++i) {
        if (visited[i] == false) {
            _dfs(i, visited);
            OutputNewComponent(visited);
        }
    }
}
void Graph::OutputNewComponent(const Array<bool> &visited) const {
    std::cout << "OutputNewComponent: ";
    for (int i = 0, len = visited.count(); i < len; ++i)
        if (visited[i]) std::cout << " -> " << i;
    std::cout << std::endl;
}



void Graph::kruskal() const {
    /////////////////////////////////////////////////////////////////
    // 1. edge cost    
    Vector<Edge> edge_list;
    for (int i = 0, vCount = _vertices.count(); i < vCount; ++i) {
        const std::vector<Edge> &adjList = _vertices[i].adjacencyList();
        for (int j = 0, adjCount = adjList.size(); j < adjCount; ++j) {
            const Edge &edge = adjList[j];
            if (edge_list.exist(edge) == false)
                edge_list.insert(edge);
        }
    }
    PriorityQueue<Edge> edges(compareEdge);
    for (int i = 0, eCount = edge_list.count(); i < eCount; ++i) {
        edges.enqueue(edge_list[i]);
    }

    // initializing ended
    /////////////////////////////////////////////////////////////////



    /////////////////////////////////////////////////////////////////
    // 2. 
    const int n = _vertices.count();
    int edge_count = 0;
    Graph tree(n);
    while (edge_count < n - 1 && edges.empty() == false) {
        Edge &&edge = edges.dequeue();
        std::cout << edge.source() << " -> " << edge.destination() << " (" << edge.cost() << ")" << std::endl;

        if (tree._path_exist(edge.source(), edge.destination()) == false) {
            tree.link(edge.source(), edge.destination(), edge.cost());
            ++edge_count;
        }
    }
    if (edge_count != n - 1) {
        std::cout << "no spanning tree" << std::endl;
        return;
    }


    tree.setAction(show_visited_vertex);
    tree.dfs(0);
    tree.setAction(nullptr);

    // 
    /////////////////////////////////////////////////////////////////
}


void Graph::prim() const {
    // assume that G has at least one vertex.
    const int n = _vertices.count();
    Graph tree(n);



    Array<bool> visited(n);
    for (int i = 0; i < n; ++i) {
        visited[i] = false;
    }
    visited[0] = true;



    int edge_count = 0;
    for (int i = 0; edge_count < n - 1; ++i) {
        int min_cost = INT_MAX;
        int src = -1, dst = -1;

        for (int j = 0; j < n; ++j) {
            if (visited[j] == false)
                continue;

            const Vertex &vertex = _vertices[j];
            const std::vector<Edge> &adjList = vertex.adjacencyList();
            for (int k = 0, adjCount = adjList.size(); k < adjCount; ++k) {
                const Edge &edge = adjList[k];
                // check src/dst included
                if (visited[edge.destination()])
                    continue;
                // check cost is minimum
                else if (min_cost > edge.cost()) {
                    src = edge.source();
                    dst = edge.destination();
                    min_cost = edge.cost();
                }
            }
        }
        if (src < 0)
            break; // there's no such edge

        tree.link(src, dst, min_cost);
        visited[dst] = true;
        ++edge_count;
    }
    if (edge_count != n - 1) {
        std::cout << "no spanning tree" << std::endl;
        return;
    }

    tree.setAction(show_visited_vertex);
    tree.dfs(0);
    tree.setAction(nullptr);
}



void Graph::init_length() {
    const int n = _vertices.count();
    _length = new Array2D<int>(n, n);
    for (int r = 0; r < n; ++r) {
        for (int c = 0; c < n; ++c) {
            if (r == c)
                _length->set(0, r, c);
            else
                _length->set(DIST_INF, r, c);
        }
    }
    for (int i = 0; i < n; ++i) {
        const std::vector<Edge> &adjList = _vertices[i].adjacencyList();
        for (int j = 0, adjCount = adjList.size(); j < adjCount; ++j) {
            const Edge &edge = adjList[j];
            _length->set(edge.cost(), edge.source(), edge.destination());
        }
    }
}
void Graph::dijkstra(int v) const {
    const int n = _vertices.count();
    const Array2D<int> &length = *_length;
    Array<bool> sp_found(n);
    Array<int> distance(n);

    for (int i = 0; i < n; ++i) {
        sp_found[i] = false;
        distance[i] = length.at(v, i);
    }
    sp_found[v] = true;

    for (int i = 0; i < n - 2; ++i) {
        int u = choose_min_dist(distance, sp_found);
        if (u < 0)
            break; // dijkstra end
        sp_found[u] = true;

        for (int w = 0; w < n; ++w) {
            if (sp_found[w] == false
                && distance[w] > distance[u] + length.at(u, w)) {
                distance[w] = distance[u] + length.at(u, w);
            }
        }
    }

    for (int i = 0; i < n; ++i) {
        std::cout << '\t' << distance[i];
    }
    std::cout << std::endl;
}



void Graph::_dfs(int v, Array<bool> &visited) const {
    _visit(v, visited);
    const std::vector<Edge> &adjList = _vertices[v].adjacencyList();
    for (int i = 0, len = adjList.size(); i < len; ++i)
        if (visited[adjList[i].destination()] == false) _dfs(adjList[i].destination(), visited);
}
void Graph::_bfs(int v, Array<bool> &visited) const {
    std::queue<int> q;
    _visit(v, visited);
    q.push(v);
    while (q.empty() == false) {
        v = q.front();
        const std::vector<Edge> &adjList = _vertices[v].adjacencyList();
        q.pop();
        for (int i = 0, count = adjList.size(); i < count; ++i) {
            int w = adjList[i].destination();
            if (visited[w] == false) {
                _visit(w, visited);
                q.push(w);
            }
        }
    }
}



void Graph::_visit(int v, Array<bool> &visited) const {
    visited[v] = true;

    if (_action) {
        _action(v);
    }
}
bool Graph::_path_exist(int src, int dst) const {
    const int n = _vertices.count();
    Array<bool> visited(n);
    for (int i = 0; i < n; ++i)
        visited[i] = false;
    _dfs(src, visited);
    return visited[dst];
}
int Graph::choose_min_dist(const Array<int> &dist, const Array<bool> &sp_found) const {
    const int n = dist.count();
    int min_cost = INT_MAX, min_idx = -1;
    for (int i = 0; i < n; ++i) {
        if (sp_found[i] == false && dist[i] < min_cost) {
            min_cost = dist[i];
            min_idx = i;
        }
    }
    return min_idx;
}