#ifndef __GRAPH_H__
#define __GRAPH_H__

#include "Array.h"
#include "Array2D.h"
#include "Edge.h"
#include "Vertex.h"

#include <iostream>
#include <queue>
#include "PriorityQueue.h"
#include "Dictionary.h"

const int DIST_INF = 9999;



class Graph {
    Array<Vertex> _vertices;
    void(*_action)(int v);

    Array2D<int> *_length;

public:
    Graph(int vertices_count);
    ~Graph();

    void setAction(void(*action)(int v));



    void dilink(int v1, int v2, int cost);
    void link(int v1, int v2, int cost);



    void dfs(int v) const;
    void bfs(int v) const;

    void Components() const;
    void OutputNewComponent(const Array<bool> &visited) const;

    void kruskal() const;
    void prim() const;

    void init_length();
    void dijkstra(int v) const;

    void floyd() const {
        const int n = _vertices.count();

        Array2D<int> length = *_length;
        std::cout << "A^(" << -1 << ")" << std::endl;
        show(length);
        std::cout << std::endl;

        for (int k = 0; k < n; ++k) {
            for (int i = 0; i < n; ++i) {
                for (int j = 0; j < n; ++j) {
                    if (length.at(i, j)>length.at(i, k) + length.at(k, j))
                        length.set(length.at(i, k) + length.at(k, j), i, j);
                }
            }
            std::cout << "A^(" << k << ")" << std::endl;
            show(length);
            std::cout << std::endl;
        }
    }
    void show(const Array2D<int> &length) const {
        const int n = _vertices.count();
        for (int i = 0; i < n; ++i) {
            for (int j = 0; j < n; ++j) {
                std::cout << '\t' << length.at(i, j);
            }
            std::cout << std::endl;
        }
    }



private:
    void _dfs(int v, Array<bool> &visited) const;
    void _bfs(int v, Array<bool> &visited) const;
    void _visit(int v, Array<bool> &visited) const;

    bool _path_exist(int src, int dst) const;

    int choose_min_dist(const Array<int> &dist, const Array<bool> &sp_found) const;
};

#endif