//////////////////////////////////////////////////////////////
// File Name : PriorityQueue.h
// Date : 2015/12/02
// Compiler : g++ 4.6.3
// Os : ubuntu 12.04.5
// Author : Do Yeong Han
// Student ID : 2011722086
// =========================================================
// ver : 2.0.7
// Description : Priority Queue data structure
// Etc.. : Data Structure Assignment #3
//  ( Link State Routing Simulation )
//////////////////////////////////////////////////////////////
#ifndef __HANDY_PRIORITY_QUEUE_H__
#define __HANDY_PRIORITY_QUEUE_H__

#include "handy.h"



template <class T>
class PriorityQueue { // NOT JUST MINPQ; it is for general purpose
    ////////////////////////////////////////////////////////////////////
    // auxilary class definition
    class PriorityQueueNode {
        T _data;
        PriorityQueueNode *_left;
        PriorityQueueNode *_right;
        int _shortest;

    public:
        ////////////////////////////////////////////////////////////////////
        // PriorityQueueNode::PriorityQueueNode
        // =================================================================
        // Input:
        //  none
        // Purpose: 
        //  constructor
        ////////////////////////////////////////////////////////////////////
        PriorityQueueNode() : _left(null), _right(null), _shortest(1) {

        }
        ////////////////////////////////////////////////////////////////////
        // PriorityQueueNode::PriorityQueueNode
        // =================================================================
        // Input:
        //  (const T& data) -> data of node
        // Purpose: 
        //  constructor
        ////////////////////////////////////////////////////////////////////
        PriorityQueueNode(const T& data) : _data(data), _left(null), _right(null), _shortest(1) {

        }
        ////////////////////////////////////////////////////////////////////
        // PriorityQueueNode::~PriorityQueueNode
        // =================================================================
        // Purpose: 
        //  destructor
        ////////////////////////////////////////////////////////////////////
        ~PriorityQueueNode() {

        }



        ////////////////////////////////////////////////////////////////////
        // PriorityQueueNode::data
        // =================================================================
        // Input:
        //  none
        // Output:
        //  (T&) -> reference to data of node
        // Purpose: 
        //  getter method for field _data
        ////////////////////////////////////////////////////////////////////
        T& data() {
            return _data;
        }
        ////////////////////////////////////////////////////////////////////
        // PriorityQueueNode::data
        // =================================================================
        // Input:
        //  none
        // Output:
        //  (const T&) -> constant reference to data of node
        // Purpose: 
        //  getter method for field _data
        ////////////////////////////////////////////////////////////////////
        const T& data() const {
            return _data;
        }
        ////////////////////////////////////////////////////////////////////
        // PriorityQueueNode::setData
        // =================================================================
        // Input:
        //  (const T& value) -> new data of node
        // Output:
        //  none
        // Purpose: 
        //  setter method for field _data
        ////////////////////////////////////////////////////////////////////
        void setData(const T &value) {
            _data = value;
        }



        ////////////////////////////////////////////////////////////////////
        // PriorityQueueNode::left
        // =================================================================
        // Input:
        //  none
        // Output:
        //  (PriorityQueueNode *) -> address of left child node
        // Purpose: 
        //  getter method for field _left
        ////////////////////////////////////////////////////////////////////
        PriorityQueueNode *left() const {
            return _left;
        }
        ////////////////////////////////////////////////////////////////////
        // PriorityQueueNode::setLeft
        // =================================================================
        // Input:
        //  (PriorityQueueNode *value) -> new address of left child node
        // Output:
        //  none
        // Purpose: 
        //  setter method for field _left
        ////////////////////////////////////////////////////////////////////
        void setLeft(PriorityQueueNode *value) {
            _left = value;
        }
        ////////////////////////////////////////////////////////////////////
        // PriorityQueueNode::right
        // =================================================================
        // Input:
        //  none
        // Output:
        //  (PriorityQueueNode *) -> address of right child node
        // Purpose: 
        //  getter method for field _right
        ////////////////////////////////////////////////////////////////////
        PriorityQueueNode *right() const {
            return _right;
        }
        ////////////////////////////////////////////////////////////////////
        // PriorityQueueNode::setRight
        // =================================================================
        // Input:
        //  (PriorityQueueNode *value) -> new address of right child node
        // Output:
        //  none
        // Purpose: 
        //  setter method for field _right
        ////////////////////////////////////////////////////////////////////
        void setRight(PriorityQueueNode *value) {
            _right = value;
        }
        ////////////////////////////////////////////////////////////////////
        // PriorityQueueNode::shortest
        // =================================================================
        // Input:
        //  none
        // Output:
        //  (int) -> shortest value
        // Purpose: 
        //  getter method for field _shortest
        ////////////////////////////////////////////////////////////////////
        int shortest() const {
            return _shortest;
        }
        ////////////////////////////////////////////////////////////////////
        // PriorityQueueNode::setShortest
        // =================================================================
        // Input:
        //  (int value) -> new shortest value to set
        // Output:
        //  none
        // Purpose: 
        //  setter method for field _shortest
        ////////////////////////////////////////////////////////////////////
        void setShortest(int value) {
            _shortest = value;
        }
    };

    // end of auxilary class definition
    ////////////////////////////////////////////////////////////////////



    ////////////////////////////////////////////////////////////////////
    // field definition
    PriorityQueueNode *_root;
    int(*_compare)(const T& data1, const T& data2); // for comparison

    // end of field definition
    ////////////////////////////////////////////////////////////////////



public:
    ////////////////////////////////////////////////////////////////////
    // PriorityQueue::PriorityQueue
    // =================================================================
    // Input:
    //  (int (*comp)(const T&, const T&)) -> comparer to manage PriorityQueue
    //                                      define comparer function first
    // *** NOTICE ***
    //  int (*comp)(const T& d1, const T& d2): *** HIGHER priority comes out first ***
    //   high priority element must be dequeued out BEFORE lower one does
    //   -1 : d1's priority is lower than d2's one  (d1<d2) -> need to swap
    //   0  : d1 and d2 has same priority elements  (d1=d2)
    //   1  : d1's priority is higher than d2's one (d1>d2)
    // Purpose: 
    //  constructor
    ////////////////////////////////////////////////////////////////////
    PriorityQueue(int (*comp)(const T&, const T&)) : _root(null), _compare(comp) {

    }
    ////////////////////////////////////////////////////////////////////
    // PriorityQueue::~PriorityQueue
    // =================================================================
    // Purpose: 
    //  destructor
    ////////////////////////////////////////////////////////////////////
    ~PriorityQueue() {
        _deallocate(_root);
    }



    ////////////////////////////////////////////////////////////////////
    // PriorityQueue::empty
    // =================================================================
    // Input:
    //  none
    // Output:
    //  (bool) -> true when queue is empty
    // Purpose: 
    //  check queue is empty
    ////////////////////////////////////////////////////////////////////
    bool empty() const {
        return _root == null;
    }



    ////////////////////////////////////////////////////////////////////
    // PriorityQueue::enqueue
    // =================================================================
    // Input:
    //  (const T& data) -> new data to insert
    // Output:
    //  none
    // Purpose: 
    //  enqueue data to priority queue
    ////////////////////////////////////////////////////////////////////
    void enqueue(const T& data) {
        /// enqueue new huffman tree ///////////////////////////////////
        // create new node
        PriorityQueueNode *newNode = new PriorityQueueNode(data); // create new node
        if (empty()) { // queue is empty
            _root = newNode; // root becomes new node
        }
        else { // queue is not empty
            _root = _meld(_root, newNode); // meld root and new node
        }

        /// end of enqueue /////////////////////////////////////////////
    }
    ////////////////////////////////////////////////////////////////////
    // PriorityQueue::dequeue
    // =================================================================
    // Input:
    //  none
    // Output:
    //  (T) -> dequeued data
    // Purpose: 
    //  dequeue data from priority queue
    ////////////////////////////////////////////////////////////////////
    T dequeue() {
        /// dequeue minimum priority huffman tree //////////////////////
        if (empty()) // cannot dequeue from empty priority queue;
            throw Exception("cannot dequeue from epmty priority queue"); // message ignored

        PriorityQueueNode *del_node = _root; // preserve node to delete
        T ret = _root->data(); // preserve data of node to delete
        _root = _meld(_root->left(), _root->right()); // meld root's left and right
        delete del_node; // it avoids del_node's destructor to deallocate it's _huffmanTree field
        return ret; // return deleted nodes huffman tree

        /// end of dequeue /////////////////////////////////////////////
    }



private:
    ////////////////////////////////////////////////////////////////////
    // PriorityQueue::_deallocate
    // =================================================================
    // Input:
    //  (PriorityQueueNode *node) -> node to deallocate
    // Output:
    //  none
    // Purpose: 
    //  deallocate all nodes in tree
    ////////////////////////////////////////////////////////////////////
    void _deallocate(PriorityQueueNode *node) {
        /// deallocate all nodes in tree ///////////////////////////////
        if (node == null) // walker is null
            return; // tree which has this node as root is empty; exit

        // deallocate children first
        _deallocate(node->left());
        _deallocate(node->right());

        // deallocate itself
        delete node;

        /// end of deallocation ////////////////////////////////////////
    }
    ////////////////////////////////////////////////////////////////////
    // PriorityQueue::_meld
    // =================================================================
    // Input:
    //  (PriorityQueueNode *a) -> tree to meld
    //  (PriorityQueueNode *b) -> tree to meld
    // Output:
    //  (PriorityQueueNode *) -> root node of tree resulted from melding
    //                           when either a or b is null, return a or b which is NOT null
    //                           null when both a and b is null
    // Purpose: 
    //  meld to tree and make new tree
    ////////////////////////////////////////////////////////////////////
    PriorityQueueNode *_meld(PriorityQueueNode *a, PriorityQueueNode *b) {
        /// meld to tree and make new tree /////////////////////////////
        if (a == null || b == null) { // either a or b is null
            // check a is NOT null; if not, return a. else, return b.
            // it does not matter whether b is null
            // (b != null: b, b == null: null == b)
            return a ? a : b;
        }

        // compare priority of node A and node B
        if (_compare(b->data(), a->data()) < 0) {
            // make node pointer A points node which has higher priority (minimum value)
            //      node pointer B points node which has lower priority
            // therefore, need to swap pointer A and pointer B
            PriorityQueueNode *tmp = a;
            a = b;
            b = tmp;
        }

        if (a->right() == null) { // right is empty
            // set B as right child of A
            a->setRight(b);
        }
        else { // right is NOT empty
            // set (melding result of A.right and B) as right child of A
            a->setRight(_meld(a->right(), b));
        }

        // check tree that it's children has appropriate position by priority
        if (a->left() == null // right exists but left does not exist
            || a->left()->shortest() < a->right()->shortest()) { // right tree is longer that left
            // swap children's position
            PriorityQueueNode *tmp = a->left();
            a->setLeft(a->right());
            a->setRight(tmp);
        }

        // set parent's shortest value
        if (a->right() == null) { // right is null; parent's shortest value = 1
            a->setShortest(1);
        }
        else { // right is not null; parent's shortest value = parent.right.shortest + 1;
            a->setShortest(a->right()->shortest() + 1);
        }

        return a; // return new tree's root node address
        /// end of melding /////////////////////////////////////////////
    }
};

#endif